#!/bin/sh

# Testcase for Overlay FileSystem (OverlayFS or short: ovl)
#
# Original Testcase: http://livenet.selfip.com/ftp/debian/overlayfs/test/test06.sh
#   Original Author: Jordi Pujol <jordipujolp@gmail.com>
#
# Slightly modified: Sedat Dilek <sedat.dilek@gmail.com>
#
# 16-Aug-2012: v4
# * Retested with OverlayFS v14 against Linux-Next (next-20120816)
# * Added a missing comment sign ("#")
# 30-Apr-2012: v3
# * Retested with OverlayFS v12 against Linux-3.4-rc5
# * Patches: 1. ovl-patch (OpenWrt) 2. my include-fix 3. my touch_atime-fix 4. d_make_root-fix from quantal (by apw)
# * apw asked for inclusion of my "fs: Increase limit filesystem stacking depth to 3" patch
#   (dmesg: [  476.709418] overlayfs: maximum fs stacking depth exceeded)
# 29-Apr-2012: v2
# * Retested with OverlayFS v12 against Linux-3.3.4
# 22-Jul-2011: v1
# * First tests with OverlayFS v10 on top of vfs-next with linux-next (next-20110722)
#   (Requires "fs: Increase limit filesystem stacking depth to 3" patch)
# 

set -x
set -e

export LANG=C
export LC_ALL=C

BASE_DIR=$(pwd)
TEMP_DIR="${BASE_DIR}/tmp"

KVER=$(uname -r)

FS_TYPE="ext4"
FS_OPTS="-F -q"

MKFS_BIN="mkfs.${FS_TYPE}"

# Load loop and overlayfs kernel-modules
modprobe -v loop
modprobe -v overlayfs

# file or dir mode, computed with current umask
# param: less restrictive mode
_mode() {
	printf "%#o" $((0${1} & ~0$(umask)))
}

# create a temporary file
# param: directory
# param: file name
_mktemp() {
	local d
	d="$(mktemp -p "${1}" ${2:+"${2}-XXX"})" || return 1
	chmod $(_mode "0666") "${d}"
	echo "${d}"
}

# create a temporary directory
# param: directory
# param: directory name
_mktempdir() {
	local d
	d="$(mktemp -d -p "${1}" ${2:+"${2}-XXX"})" || return 1
	chmod $(_mode "0777") "${d}"
	echo "${d}"
}

_sample_data() {
	local p="${1}"
	local l="$(_mktempdir "${WORK_DIR}" "WORK")"
	local d="${l}${p}"
	local f
	( mkdir -p "${d}"
	echo "a" > "$(_mktemp "${d}" "TEXT")"
	echo "a" > "$(_mktemp "${d}" "TEXT")"
	echo "a" > "$(_mktemp "${d}" "TEXT")"
	f="$(_mktemp "${d}" "EXEC")"
	echo "#/bin/sh
echo 'executed'
:" > "${f}"
	chmod $(_mode "0777") "${f}"
	mksquashfs "${l}" "${l}.squashfs" -b 1M
	rm -rf "${l}" ) >&2
	echo "${l}.squashfs"
}

_overlayfs_test() {
	local d="${1}"
	local rc="0"
	local f g i

	ls -la "${d}"

	i=0
	g="${d}"
	while [ ${i} -lt 3 ]; do
		if g="$(_mktempdir "${g}" "TMPDIR")"; then
			_mktemp "${g}" "TMPFILE"
			i=$((${i}+1))
			rm ${g}/*
		else
			break
		fi
	done

	for f in $(find "${d}" -type f); do
		sed -i -e 's|a|&b\n&&|g' "${f}" || rc="1"
	done

	for f in $(find "${d}" -type f -name 'EXEC*'); do
		"${f}" || echo "not executed"
	done

	if [ "${rc}" != "0" ]; then
		for f in $(find "${d}" -type f); do
			touch "${f}" || :
		done

		for f in $(find "${d}" -type f); do
			sed -i -e 's|a|a|g' "${f}" || :
		done
	fi

	for f in $(find "${d}" -type f); do
		cat "${f}"
		echo "--------"
	done
}

_test() {
	_overlayfs_test "${ROOT_DIR}/etc/apache2"

	_overlayfs_test "${ROOT_DIR}/boot/grub"

	_overlayfs_test "${ROOT_DIR}/etc/default"
}

[ -d ${TEMP_DIR} ] || mkdir -p ${TEMP_DIR}

WORK_DIR="$(_mktempdir "${TEMP_DIR}" "TEST-${KVER}")"

LOG="$(_mktemp "${WORK_DIR}" "LOG")"
exec > "${LOG}" 2>&1

LOWER1_DIR="$(_sample_data "/etc/default")"
LOWER2_DIR="$(_sample_data "/boot/grub")"
LOWER3_DIR="$(_sample_data "/etc/apache2")"

unset uf lf
while read uf; do
	[ -e "${uf}" ] || \
		continue

	if [ -d "${uf}" ]; then
		u="${uf}"
	else
		u="$(_mktempdir "${WORK_DIR}" "UPPER")"
		mount -o loop,ro "${uf}" "${u}"
	fi

	if [ -n "${lf}" ]; then
		if [ -d "${lf}" ]; then
			l="${lf}"
		else
			l="$(_mktempdir "${WORK_DIR}" "LOWER")"
			mount -o loop,ro "${lf}" "${l}"
		fi

		r="$(_mktempdir "${WORK_DIR}" "ROOT-RO")"

		mount -t overlayfs -o ro,lowerdir="${l}",upperdir="${u}" "overlayfs" "${r}"

		lf="${r}"
	else
		lf="${u}"
	fi
done << EOF
${LOWER1_DIR}
${LOWER2_DIR}
${LOWER3_DIR}
EOF

[ -d "${lf}" ] || \
	exit 1

ROOT_DIR="$(_mktempdir "${WORK_DIR}" "ROOT")"
COW_DIR="$(_mktempdir "${WORK_DIR}" "COW")"
#mount -t tmpfs -o rw tmpfs "${COW_DIR}"
COW_FILE="$(_mktemp "${WORK_DIR}" "COWFILE")"
dd if=/dev/zero of=${COW_FILE} bs=1024 count=0 seek=$((1024*128))
${MKFS_BIN} ${FS_OPTS} "${COW_FILE}"
mount -o loop,rw "${COW_FILE}" "${COW_DIR}"

mount -t overlayfs -o rw,lowerdir="${lf}",upperdir="${COW_DIR}" "overlayfs" "${ROOT_DIR}"

echo "rootdir: ${ROOT_DIR}"
echo "COW: ${COW_DIR}"
mount -l

# COW
mount -o remount,ro "${ROOT_DIR}"
_test

#mount -o remount,rw "${ROOT_DIR}"
#_test

mount -o remount,rw "${COW_DIR}"
mount -o remount,rw "${ROOT_DIR}" || echo "${?}"
_test

mount -o remount,rw "${COW_DIR}"
mount -o remount,rw,noexec "${ROOT_DIR}"
_test

mount -o remount,rw,noexec "${COW_DIR}"
mount -o remount,rw,noexec "${ROOT_DIR}"
_test

# exit 0

umount "${ROOT_DIR}"

while read f; do
	[ -d "${f}" ] && \
	mountpoint -q "${f}" && \
	umount "${f}" || :
done << EOF
$(find "${WORK_DIR}" -mindepth 1 -maxdepth 1 -type d)
EOF

:
